"""pypi.

Desc: Library for getting information about Python packages by querying
      The CheeseShop (PYPI a.k.a. Python Package Index).


Author: Rob Cakebread <cakebread at gmail>

License  : BSD (See COPYING)

"""

from __future__ import print_function

import ast
import re
import platform
if platform.python_version().startswith('2'):
    import xmlrpclib
    import urllib2
else:
    import xmlrpc.client as xmlrpclib
    import urllib.request as urllib2
import os
import time

from yolk.utils import get_yolk_dir


XML_RPC_SERVER = 'https://pypi.python.org/pypi'

MAX_CACHE_AGE = 60


class ProxyTransport(xmlrpclib.Transport):

    """Provides an XMl-RPC transport routing via a http proxy.

    This is done by using urllib2, which in turn uses the environment
    varable http_proxy and whatever else it is built to use (e.g. the
    windows    registry).

    NOTE: the environment variable http_proxy should be set correctly.
    See check_proxy_setting() below.

    Written from scratch but inspired by xmlrpc_urllib_transport.py
    file from http://starship.python.net/crew/jjkunce/ by jjk.

    A. Ellerton 2006-07-06

    """

    def request(self, host, handler, request_body, verbose):
        """Send xml-rpc request using proxy."""
        # We get a traceback if we don't have this attribute:
        self.verbose = verbose
        url = 'http://' + host + handler
        request = urllib2.Request(url)
        try:
            request.add_data(request_body)
        except AttributeError:
            request.data = request_body
        # Note: 'Host' and 'Content-Length' are added automatically
        request.add_header('User-Agent', self.user_agent)
        request.add_header('Content-Type', 'text/xml')
        proxy_handler = urllib2.ProxyHandler()
        opener = urllib2.build_opener(proxy_handler)
        fhandle = opener.open(request)
        return(self.parse_response(fhandle))


def check_proxy_setting():
    """If the environmental variable 'HTTP_PROXY' is set, it will most likely
    be in one of these forms:

    proxyhost:8080 http://proxyhost:8080 urlllib2
    requires the proxy URL to start with 'http://' This routine does that, and
    returns the transport for xmlrpc.

    """
    try:
        http_proxy = os.environ['HTTP_PROXY']
    except KeyError:
        return

    if not http_proxy.startswith('http://'):
        match = re.match(r'(http://)?([-_\.A-Za-z]+):(\d+)', http_proxy)
        os.environ['HTTP_PROXY'] = 'http://%s:%s' % (match.group(2),
                                                     match.group(3))
    return


class CheeseShop(object):

    """Interface to Python Package Index."""

    def __init__(self, debug=False, no_cache=False, yolk_dir=None):
        self.debug = debug
        self.no_cache = no_cache
        if yolk_dir:
            self.yolk_dir = yolk_dir
        else:
            self.yolk_dir = get_yolk_dir()
        self.xmlrpc = self.get_xmlrpc_server()
        self.pkg_cache_file = self.get_pkg_cache_file()
        self.pkg_list = None
        self.get_cache()

    def get_cache(self):
        """Get a package name list from disk cache or PyPI."""
        # This is used by external programs that import `CheeseShop` and don't
        # want a cache file written to ~/.pypi and query PyPI every time.
        if self.no_cache:
            self.pkg_list = self.list_packages()
            return

        if not os.path.exists(self.yolk_dir):
            os.mkdir(self.yolk_dir)
        try:
            self.pkg_list = self.query_cached_package_list()
        except (IOError, ValueError):
            self.fetch_pkg_list(True)

    def get_xmlrpc_server(self):
        """Returns PyPI's XML-RPC server instance."""
        check_proxy_setting()
        if 'XMLRPC_DEBUG' in os.environ:
            debug = 1
        else:
            debug = 0
        return xmlrpclib.Server(XML_RPC_SERVER, transport=ProxyTransport(),
                                verbose=debug)

    def get_pkg_cache_file(self):
        """Returns filename of pkg cache."""
        return os.path.abspath('%s/pkg_list.py' % self.yolk_dir)

    def query_versions_pypi(self, package_name):
        """Fetch list of available versions for a package from PyPI."""
        if package_name not in self.pkg_list:
            self.fetch_pkg_list()
        # I have to set version=[] for edge cases like "Magic file extensions"
        # but I'm not sure why this happens. It's included with Python or
        # because it has a space in it's name?
        versions = []
        for pypi_pkg in self.pkg_list:
            if normalize(pypi_pkg) == normalize(package_name):
                versions = self.package_releases(pypi_pkg)
                package_name = pypi_pkg
                break
        return (package_name, versions)

    def query_cached_package_list(self):
        """Return list of cached package names from PYPI."""
        with open(self.pkg_cache_file, 'r') as input_file:
            return ast.literal_eval(input_file.read())

    def fetch_pkg_list(self, ignore_cache=False):
        """Fetch and cache master list of package names from PYPI."""
        if (
            not ignore_cache and
            os.path.exists(self.pkg_cache_file) and
            time.time() - os.stat(self.pkg_cache_file).st_mtime < MAX_CACHE_AGE
        ):
            return

        package_list = self.list_packages()
        with open(self.pkg_cache_file, 'w') as output_file:
            print(package_list, file=output_file)
        self.pkg_list = package_list

    def search(self, spec, operator):
        """Query PYPI via XMLRPC interface using search spec."""
        return self.xmlrpc.search(spec, operator.lower())

    def changelog(self, hours):
        """Query PYPI via XMLRPC interface using search spec."""
        return self.xmlrpc.changelog(get_seconds(hours))

    def updated_releases(self, hours):
        """Query PYPI via XMLRPC interface using search spec."""
        return self.xmlrpc.updated_releases(get_seconds(hours))

    def list_packages(self):
        """Query PYPI via XMLRPC interface for a a list of all package
        names."""
        return self.xmlrpc.list_packages()

    def release_urls(self, package_name, version):
        """Query PYPI via XMLRPC interface for a pkg's available versions."""

        return self.xmlrpc.release_urls(package_name, version)

    def release_data(self, package_name, version):
        """Query PYPI via XMLRPC interface for a pkg's metadata."""
        try:
            return self.xmlrpc.release_data(package_name, version)
        except xmlrpclib.Fault:
            # XXX Raises xmlrpclib.Fault if you give non-existent version
            # Could this be server bug?
            return

    def package_releases(self, package_name):
        """Query PYPI via XMLRPC interface for a pkg's available versions."""
        return self.xmlrpc.package_releases(package_name)


def get_seconds(hours):
    """Get number of seconds since epoch from now minus `hours`

    @param hours: Number of `hours` back in time we are checking
    @type hours: int

    Return integer for number of seconds for now minus hours

    """
    return int(time.time() - (60 * 60) * hours)


def normalize(name):
    """Return normalized name."""
    return name.lower().replace('_', '-')



"""yolklib.

Desc: Library for getting information about Python packages installed by
      setuptools, package metadata, package dependencies, and querying
      The CheeseShop (PYPI) for Python package release information.


Author: Rob Cakebread <cakebread @ gmail>

License  : BSD (See COPYING)

"""

import pkg_resources


class Distributions(object):

    """Helper class for pkg_resources."""

    def __init__(self):
        """init."""

        self.environment = pkg_resources.Environment()
        self.working_set = pkg_resources.WorkingSet()

    def query_activated(self, dist):
        """Return True if distribution is active.

        @param dist: pkg_resources Distribution object

        @returns: True or False

        """
        if dist in self.working_set:
            return True
        else:
            return False

    def get_distributions(self, show, pkg_name='', version=''):
        """Yield installed packages.

        @param show: Type of package(s) to show; active, non-active or all
        @type show: string: "active", "non-active", "all"

        @param pkg_name: PyPI project name
        @type pkg_name: string

        @param version: project's PyPI version
        @type version: string

        @returns: yields tuples of distribution and True or False depending
                  on active state. e.g. (dist, True)

        """
        # pylint: disable-msg=W0612
        # "name" is a placeholder for the sorted list.
        for name, dist in self.get_alpha(show, pkg_name, version):
            ver = dist.version
            for package in self.environment[dist.project_name]:
                if ver == package.version:
                    if show == 'nonactive' and dist not in self.working_set:
                        yield (dist, self.query_activated(dist))
                    elif show == 'active' and dist in self.working_set:
                        yield (dist, self.query_activated(dist))
                    elif show == 'all':
                        yield (dist, self.query_activated(dist))

    def get_alpha(self, show, pkg_name='', version=''):
        """Return list of alphabetized packages.

        @param pkg_name: PyPI project name
        @type pkg_name: string

        @param version: project's PyPI version
        @type version: string

        @returns: Alphabetized list of tuples. Each tuple contains
                  a string and a pkg_resources Distribution object.
                  The string is the project name + version.

        """
        alpha_list = []
        for dist in self.get_packages(show):
            if pkg_name and dist.project_name != pkg_name:
                # Only checking for a single package name
                pass
            elif version and dist.version != version:
                # Only checking for a single version of a package
                pass
            else:
                alpha_list.append((dist.project_name + dist.version, dist))
        alpha_list.sort()
        return alpha_list

    def get_packages(self, show):
        """Return list of Distributions filtered by active status or all.

        @param show: Type of package(s) to show; active, non-active or all
        @type show: string: "active", "non-active", "all"

        @returns: list of pkg_resources Distribution objects

        """

        if show == 'nonactive' or show == 'all':
            all_packages = []
            for package in self.environment:
                # There may be multiple versions of same packages
                for i in range(len(self.environment[package])):
                    if self.environment[package][i]:
                        all_packages.append(self.environment[package][i])
            return all_packages
        else:
            # Only activated packages
            return self.working_set

    def case_sensitive_name(self, package_name):
        """Return case-sensitive package name given any-case package name.

        @param project_name: PyPI project name
        @type project_name: string

        """
        if len(self.environment[package_name]):
            return self.environment[package_name][0].project_name

    def get_highest_installed(self, project_name):
        """Return highest version of installed package.

        @param project_name: PyPI project name
        @type project_name: string

        @return: string of highest installed version

        """
        return self.environment[project_name][0].version


def get_highest_version(versions):
    """Returns highest available version for a package in a list of versions.

    Uses pkg_resources to parse the versions.

    @param versions: List of PyPI package versions
    @type versions: List of strings

    @returns: string of a PyPI package version

    """
    sorted_versions = []
    for ver in versions:
        sorted_versions.append((pkg_resources.parse_version(ver), ver))

    sorted_versions = sorted(sorted_versions)
    sorted_versions.reverse()
    return sorted_versions[0][1]



#!/usr/bin/env python
# pylint: disable-msg=W0231,R0904


"""setuptools_support.

License  : BSD (See COPYING)

"""


from yolk.yolklib import Distributions

from setuptools.package_index import PackageIndex
import pkg_resources


class DownloadURI(Exception):

    """Hack to raise the value of the URI from PackageIndex."""

    def __init__(self, value):
        """init."""
        self.value = value

    def __str__(self):
        """Set value to URI."""
        return repr(self.value)


class MyPackageIndex(PackageIndex):

    """Over-ride methods so we can obtain the package's URI."""

    def _download_to(self, url, filename):
        """Raise exception so we immediately get url with no downloading."""
        raise DownloadURI(url)

    def download(self, spec, tmpdir='/tmp/spambar'):
        """Raise exception so we immediately get url with no downloading."""
        raise DownloadURI(spec)


def get_download_uri(package_name, version, source, index_url=None):
    """Use setuptools to search for a package's URI.

    @returns: URI string

    """
    tmpdir = None
    force_scan = True
    develop_ok = False
    if not index_url:
        index_url = 'https://pypi.python.org/pypi'

    if version:
        pkg_spec = '%s==%s' % (package_name, version)
    else:
        pkg_spec = package_name
    req = pkg_resources.Requirement.parse(pkg_spec)
    pkg_index = MyPackageIndex(index_url)
    try:
        pkg_index.fetch_distribution(req, tmpdir, force_scan, source,
                                     develop_ok)
    except DownloadURI as url:
        # Remove #egg=pkg-dev
        clean_url = url.value.split('#')[0]
        # If setuptools is asked for an egg and there isn't one, it will
        # return source if available, which we don't want.
        if not source and not clean_url.endswith('.egg') and \
                not clean_url.endswith('.EGG'):
            return
        else:
            return clean_url


def get_pkglist():
    """Return list of all installed packages.

    Note: It returns one project name per pkg no matter how many versions
    of a particular package is installed

    @returns: list of project name strings for every installed pkg

    """

    dists = Distributions()
    projects = []
    for (dist, _) in dists.get_distributions('all'):
        if dist.project_name not in projects:
            projects.append(dist.project_name)
    return projects



"""Miscellaneous functions.

run_command borrowed from Cheesecake - See CREDITS.

"""

import os
import shlex
import signal
import subprocess
import time


def get_yolk_dir():
    """Return location we store config files and data."""
    return os.path.abspath('%s/.yolk' % os.path.expanduser('~'))


def run_command(cmd, env=None, max_timeout=None):
    r"""Run command and return its return status code and its output.

    >>> run_command('true')
    (0, '')

    >>> run_command('false')
    (1, '')

    >>> run_command('echo hello world')
    (0, 'hello world\n')

    """
    try:
        pipe = subprocess.Popen(shlex.split(cmd),
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                env=env)
    except Exception as errmsg:
        return 1, errmsg

    # Wait only max_timeout seconds.
    if max_timeout:
        start = time.time()
        while pipe.poll() is None:
            time.sleep(0.1)
            if time.time() - start > max_timeout:
                os.kill(pipe.pid, signal.SIGINT)
                pipe.wait()
                return 1, 'Time exceeded'

    pipe.wait()
    return pipe.returncode, pipe.communicate()[0].decode()


def command_successful(cmd):
    """Returns True if command exited normally, False otherwise."""
    return_code, _ = run_command(cmd)
    return return_code == 0



#!/usr/bin/env python

"""Installer for yolk."""

import ast

from setuptools import setup


def version():
    """Return version string."""
    with open('yolk/__init__.py') as input_file:
        for line in input_file:
            if line.startswith('__version__'):
                return ast.parse(line).body[0].value.s


with open('README.rst') as readme:
    setup(
        name='yolk3k',
        license='BSD License',
        version=version(),
        description='Command-line tool for querying PyPI and Python packages '
                    'installed on your system.',
        long_description=readme.read(),
        maintainer='Steven Myint',
        author='Rob Cakebread',
        url='https://github.com/myint/yolk',
        keywords='PyPI,setuptools,cheeseshop,distutils,eggs,package,'
                 'management',
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: BSD License',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.2',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
        packages=['yolk'],
        package_dir={'yolk': 'yolk'},
        entry_points={'console_scripts': ['yolk = yolk.cli:main']}
    )
